/* may only be used in accordance with the terms of the GNU General Public License */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include <linux/soundcard.h>

#include "charmonic.h"

/* hey, who left out PI ? */
#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef VERSION
#define VERSION "unknown"
#endif

#define C_LOWESTFREQ   2
#define C_BUFSIZE    512

#define C_RES      100UL

#define C_DSP  "/dev/dsp"

/* Clues

 * buildwave()   writes a single period of the base sound to a buffer
 * addharmonic() adds a given harmonic of given amplitude to above buffer 
 * buildout()    copies the single period buffer to output buffer N times
 *               N is large enough to occupy the entire output buffer of
 *               audio device + 1 more period, so that with mod% tricks 
 *               we can use output buffer at different offsets to write 
 *               to device
 */

struct c_entry *c_initentry(struct c_harmonic *h)
{
  struct c_entry *result;

  result = (struct c_entry *) malloc(sizeof(struct c_entry) * (1 + h->h_speed / 2));

  return result;
}

int c_initdsp(char *fname, struct c_harmonic *h)
{
  int fd;
  int format;
  int channels;
  int speed;
  int blocksize;

  fd = open(fname, O_WRONLY);

  if (fd < 0) {
    perror("initdsp(): Can't open dsp");
    return -1;
  }
#ifdef TRACE
  fprintf(stderr, "initdsp(): Opened dsp fd=<%d>\n", fd);
#endif

  ioctl(fd, SNDCTL_DSP_RESET, NULL);

#ifdef TRACE
  fprintf(stderr, "initdsp(): Reset dsp\n");
#endif

  /* 
   * I suppose big endian could be supported 
   * if one fiddles the unpacking in buildout
   */
  format = AFMT_S16_LE;
  if (ioctl(fd, SNDCTL_DSP_SETFMT, &format) < 0) {
    perror("initdsp(): Can't set sample size to le16 bits");
    close(fd);
    return -1;
  }
  if (format != AFMT_S16_LE) {
    fprintf(stderr, "Sample size of le16 bits unsupported by device\n");
    close(fd);
    return -1;
  }
#ifdef TRACE
  fprintf(stderr, "initdsp(): set samples to 16 bits\n");
#endif

  channels = 1;
  if (ioctl(fd, SNDCTL_DSP_CHANNELS, &channels) < 0) {
    perror("initdsp(): Can't set dsp to mono");
    close(fd);
    return -1;
  }
  if (channels != 1) {
    perror("initdsp(): Mono mode not supported by device");
    close(fd);
    return -1;
  }
#ifdef TRACE
  fprintf(stderr, "initdsp(): set to mono\n");
#endif

  /* Wow, my cheapo ACERMAGIC 23 supports 48kHz */
  speed = 48000;
  if (ioctl(fd, SNDCTL_DSP_SPEED, &speed) < 0) {
    perror("initdsp(): Can't set dsp speed");
    close(fd);
    return -1;
  }
  h->h_speed = speed;

#ifdef TRACE
  fprintf(stderr, "initdsp(): Best speed is <%d>\n", h->h_speed);
#endif

  if (ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &blocksize) < 0) {
    perror("initdsp(): Can't establish dsp buffer size");
    close(fd);
    return -1;
  }
  h->h_chunk = (int) blocksize;

#ifdef TRACE
  fprintf(stderr, "initdsp(): dsp blocksize=<%d>\n", h->h_chunk);
#endif

  return fd;
}

int c_initharmonic(struct c_harmonic *h)
{

  /* the single wave buffer. No point in fighting Nyquist */
  h->h_wavelen = h->h_speed / C_LOWESTFREQ;
  h->h_wave = (signed short *) malloc(sizeof(signed short) * (h->h_wavelen));
  if (h->h_wave == NULL) {
    perror("initharmonic(): Can not allocate buffer");
    return -1;
  }
  /* the funky output buffer with %offset */
  h->h_coffset = 0;
  h->h_out = (unsigned char *) malloc(sizeof(signed short) * h->h_wavelen + sizeof(unsigned char) * h->h_chunk);
  if (h->h_out == NULL) {
    perror("initharmonic(): Can not allocate buffer");
    return -1;
  }
  h->h_wavelen = 0;

  return 0;
}

int c_buildwave(int samples, int amplitude, struct c_harmonic *h)
{
  int i;

  h->h_wavelen = samples;

  for (i = 0; i < h->h_wavelen; i++) {
    h->h_wave[i] = (signed short) (sin((2.0 * PI * (float) i) / ((float) samples)) * (double) amplitude);
#ifdef TRACELOTS
    fprintf(stderr, "buildwave(): <%d>,<%d>\n", i, h->h_wave[i - 1]);
#endif
  }

  return 0;
}

int c_addharmonic(int monic, int amplitude, struct c_harmonic *h)
{
  int i;

  for (i = 0; i < h->h_wavelen; i++) {
    h->h_wave[i] += (signed short) (sin((2.0 * PI * (float) (i * (1 + monic))) / (float) (h->h_wavelen)) * (double) amplitude);
#ifdef TRACE
    fprintf(stderr, "addharmonic(): <%d>,<%d>\n", i, h->h_wave[i - 1]);
#endif
  }

  return 0;
}

int c_clearend(struct c_harmonic *h)
{
  int wait, i, s, l;

  /* insert a period of silence but only after sin(360)=0 to avoid big click */
  if (h->h_coffset) {
    wait = 1;
    i = (2 * h->h_wavelen) - (h->h_coffset);
    while (i > h->h_chunk) {
      i -= h->h_chunk;
    }
    /* length of zeroing out */
    l = h->h_chunk - i;
    /* start of zeroing out */
    s = (2 * h->h_wavelen);

#ifdef TRACE
    fprintf(stderr, "clearend(): wait=<%d>\n", wait);
#endif

    memset(h->h_out + s, 0, l);

#ifdef TRACE
    fprintf(stderr, "clearend(): Zeroed <%d> from <%d>\n", l, s);
#endif

  } else {
    wait = 0;
  }

  return wait;
}

int c_buildout(struct c_harmonic *h)
{
  int i, j;

  h->h_coffset = 0;

  /* for big endian this should do a different unpacking */
  for (i = 0, j = 0; i < (h->h_chunk + (2 * h->h_wavelen)); i++, j += 2) {
    h->h_out[j] = (unsigned char) (h->h_wave[i % h->h_wavelen] & 0xff);
    h->h_out[j + 1] = (unsigned char) (((h->h_wave[i % h->h_wavelen]) >> 8) & 0xff);
#ifdef TRACELOTS
    fprintf(stderr, "buildout(): <%d>,<%d><%d>\n", i, h->h_out[j], h->h_out[j + 1]);
#endif
  }
  return 0;
}

int c_playout(int fd, struct c_harmonic *h)
{
  int t;

  t = write(fd, h->h_out + h->h_coffset, h->h_chunk);
#ifdef TRACERAW
  write(STDOUT_FILENO, h->h_out + h->h_coffset, h->h_chunk);
#endif

  if (t > 0) {
#ifdef TRACEPOLL
    fprintf(stderr, "playout(): Wrote <%d> from <%d> to <%d>", t, h->h_coffset, h->h_coffset + h->h_chunk);
#endif
    h->h_coffset = (h->h_coffset + t) % (h->h_wavelen * 2);
#ifdef TRACEPOLL
    fprintf(stderr, " new offset is <%d>\n", h->h_coffset);
#endif
  } else {
    perror("playout(): write failed");
  }

  return 0;
}

int c_dumpwave(struct c_harmonic *h, char *fname)
{
  int i = 0;
  FILE *fp;
  int result = 0;

  if (strcmp(fname, "-") && (fname[0] != '\0')) {
    fp = fopen(fname, "w");
    if (fp == NULL) {
      perror("dumpwave(): Can not open file");
      result = (-1);
    } else {
      for (i = 0; i < h->h_wavelen; i++) {
	fprintf(fp, "% 06d\n", h->h_wave[i]);
      }
      fclose(fp);
    }
  } else {
    for (i = 0; i < h->h_wavelen; i++) {
      printf("% 06d\n", h->h_wave[i]);
    }
  }

  return 0;
}

int c_calcyc(int f, struct c_harmonic *h)
{
  int result = (-1);
  int c;

  if (f >= 4) {
    c = h->h_speed / f;
    if (c > 2) {
#ifdef TRACE
      fprintf(stderr, "calcyc(): Frequency <%d> converts to <%d> samples/cycle\n", f, c);
#endif
      result = c;
    } else {
      fprintf(stderr, "Frequency too high\n");
    }
  } else {
    fprintf(stderr, "Frequency too low\n");
  }

  return result;
}

int c_parsentry(char *s, struct c_entry *e, struct c_harmonic *h)
{
  int result = 0;
  int i, j, k, n;
  char d;
  int sum;

  i = 0;
  j = 0;
  while (isdigit(s[i]))
    i++;
  d = s[i];
  s[i] = '\0';

  e[0].e_number = c_calcyc(atoi(s), h);
  if (e[0].e_number > 0) {
    e[0].e_percentage = 100;

    j = i;
    k = 1;
    n = 1;
    while (d == ':') {
      j++;
      i = 0;
      while (isdigit(s[j + i]))
	i++;
      d = s[j + i];
      if (i == 0) {
	n++;
      } else {
	s[j + i] = '\0';
	e[k].e_percentage = atoi(s + j);
	if (e[k].e_percentage < 0) {
	  fprintf(stderr, "percentage <%d> for harmonic <%d> too small\n", e[k].e_percentage, n);
	  result = (-1);
	  d = '\n';
	} else if (e[k].e_percentage > 0) {
	  e[k].e_number = n;
#ifdef TRACE
	  fprintf(stderr, "parsentry(): hamonic <%d>, percentage <%d> at position <%d>\n", n, e[k].e_percentage, k);
#endif
	  if ((e[0].e_number / n) < 2) {
	    fprintf(stderr, "hamonic <%d> too high for sample rate <%d>\n", n, h->h_speed);
	    result = (-1);
	    d = '\n';
	  }
	  k++;
	  n++;
	} else {		/* percentage == 0 */
	  n++;
	}
      }
      j += i;
    }

    e[k].e_number = 0;

    if (d != '\n') {
      fprintf(stderr, "illegal character <%c> at <%d>\n", d, j + i);
      result = (-1);
    }
    if (result == 0) {
      sum = 0;
      i = 0;
      while (e[i].e_number > 0) {
	sum += e[i].e_percentage;
	i++;
      }
#ifdef TRACE
      fprintf(stderr, "parsentry(): Total <%d>\n", sum);
#endif
      i = 0;
      while (e[i].e_number > 0) {
	e[i].e_volume = (int) (((float) (e[i].e_percentage) / (float) (sum)) * 30000.0);
#ifdef TRACE
	fprintf(stderr, "parsentry(): Volume <%d> for entry <%d>\n", e[i].e_volume, i);
#endif
	i++;
      }
    }
  } else {
    result = (-1);
  }
  return result;
}

int c_do()
{
  struct c_harmonic h;
  struct c_entry *e;
  int dfd;			/* codec fd */
  int w, i;			/* tmp vars */
  char buf[C_BUFSIZE];		/* commands */
  int buft, bufo, bufs;		/* buffer total and offset */
  int stdinflags;		/* flag to flip (non)blocking */
  int run, play;		/* system state */
  int state, eol;		/* state of parser */
  int pause, loop;		/* hold tone/silence */

  dfd = c_initdsp(C_DSP, &h);
  i = c_initharmonic(&h);
  e = c_initentry(&h);

  if ((dfd >= 0) && (i == 0) && (e != NULL)) {

    run = 1;
    play = 0;
    bufo = 0;
    buft = 0;

    if (isatty(STDIN_FILENO) && isatty(STDOUT_FILENO)) {
      printf("CLI harmonic generator version %s. Type help for help\n", VERSION);
    }
    do {

      /* maybe check that buffer not too full to improve response time ? */
      if (play) {
	c_playout(dfd, &h);
      }
#ifdef TRACEPOLL
      fprintf(stderr, "do(): About to read input at offset <%d>\n", buft);
#endif
      bufs = read(STDIN_FILENO, buf + buft, C_BUFSIZE - buft);
#ifdef TRACEPOLL
      fprintf(stderr, "do(): Read input, returned <%d>\n", bufs);
#endif
      switch (bufs) {
      case -1:
	switch (errno) {
	case EINTR:
	case EAGAIN:
#ifdef TRACE
	  fprintf(stderr, "do(): E<%d>, trying again\n", errno);
#endif
	  break;
	default:
	  /* don't bother continuing */
	  bufo = 0;
	  buft = 0;
	  perror("do(): Error on reading input");
	  run = 0;
	  break;
	}
	break;
      case 0:
	/* eof : continue until end */
	run = 0;
	break;
      default:
	buft += bufs;
	break;
      }

      state = 1;
      while (state) {
	/* eat ws */
	while (isspace(buf[bufo]) && buft > bufo)
	  bufo++;
	/* something to process */
	if (buft > bufo) {
	  eol = bufo;
	  while (buf[eol] != '\n' && buf[eol] != '\r' && buft > eol)
	    eol++;
	  /* there is an end of line, so work can be done */
	  if (buft > eol) {
	    /* make sure we have a \n */
	    buf[eol] = '\n';
	    switch (buf[bufo]) {
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	      if (play) {
		w = c_clearend(&h);
		while (w > 0) {
		  c_playout(dfd, &h);
		  w--;
		}
	      }
	      if (c_parsentry(buf + bufo, e, &h)) {

		/* go blocking */
		stdinflags = fcntl(STDIN_FILENO, F_GETFL, 0);
		fcntl(STDIN_FILENO, F_SETFL, (~O_NONBLOCK) & stdinflags);

		play = 0;
	      } else {
		/* go nonblocking */
		stdinflags = fcntl(STDIN_FILENO, F_GETFL, 0);
		fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK | stdinflags);

		c_buildwave(e[0].e_number, e[0].e_volume, &h);

		i = 1;
		while (e[i].e_number > 0) {
		  c_addharmonic(e[i].e_number, e[i].e_volume, &h);
		  i++;
		}

		c_buildout(&h);
		play = 1;
	      }
	      break;
	    case 'h':
	    case 'H':
	      printf("Commands:\n");
	      printf("  <base frequency>:<integer percentage volume of first harmonic>:<etc>\n");
	      printf("  d <filename>    : dump wave to file <filename>\n");
	      printf("  echo <string>   : echo a string to stdout\n");
	      printf("  help            : this help\n");
	      printf("  quit            : end the program\n");
	      printf("  silence         : turn off sound\n");
	      printf("  version         : display version information\n");
	      printf("  #comments       : add a comment (try #!/usr/local/bin/charmonic on first line)\n");
	      printf("Example: tone A (440Hz), with a 50%% 1st harmonic as well as a 2nd, 3rd and 5th\n");
	      printf("  440:50:40:20::10\n");
	      break;
	    case 'b':
	    case 'B':
	    case 'x':
	    case 'X':
	    case 'q':
	    case 'Q':
	      /* quit */
	      fprintf(stderr, "bye\n");
	      run = 0;
	      break;
	    case 'c':
	      fprintf(stderr, "(c) 1999 by mgw - licensed under the GPL\n");
	      break;
	    case 'd':
	    case 'D':
	      /* dump the wave to file */
	      bufo++;
	      while (buf[bufo] == ' ')
		bufo++;
	      buf[eol] = '\0';
	      c_dumpwave(&h, buf + bufo);
	      break;
	    case 'e':
	    case 'E':
	      /* echo the line */
	      bufo++;
	      while (isalpha(buf[bufo]))
		bufo++;
	      while (buf[bufo] == ' ')
		bufo++;
	      for (i = bufo; i < eol; i++) {
		fputc(buf[i], stdout);
	      }
	      fputc('\n', stdout);
	      break;
	    case 'p':
	    case 'P':
	      bufo++;
	      while (buf[bufo] == ' ' || isalpha(buf[bufo]))
		bufo++;
	      pause = atoi(buf + bufo);
	      if (pause > 0) {
		if (play) {
		  if (h.h_wavelen > 0) {
		    loop = ((h.h_speed * pause) / (C_RES * h.h_chunk));
#ifdef TRACE
		    fprintf(stderr, "do(): Pausing for %d/%lu seconds = %d periods of %d samples at %d Hz\n", pause, C_RES, loop, h.h_chunk, h.h_speed);
#endif
		    while (loop > 0) {
		      c_playout(dfd, &h);
		      loop--;
		    }
		  } else {
		    fprintf(stderr, "Eh, wave is zero ?\n");
		  }
		} else {
		  usleep(pause * (1000000 / C_RES));
		}
	      } else {
		fprintf(stderr, "Pause not specified correctly\n");
	      }
	      break;
	    case 's':
	    case 'S':
#ifdef TRACE
	      fprintf(stderr, "do(): Turning on silence\n");
#endif
	      /* silence */
	      if (play) {
		w = c_clearend(&h);
		while (w > 0) {
		  c_playout(dfd, &h);
		  w--;
		}
		/* go blocking */
		stdinflags = fcntl(STDIN_FILENO, F_GETFL, 0);
		fcntl(STDIN_FILENO, F_SETFL, (~O_NONBLOCK) & stdinflags);
		play = 0;
	      }
	      break;
	    case '#':
	      /* ignore comments */
	      break;
	    case 'v':
	      /* print version */
	      printf("Charmonic: CLI tone generator version %s\n", VERSION);
	      break;
	    default:
	      /* complain */
	      fprintf(stderr, "Unknown command <");
	      for (i = bufo; i < eol; i++) {
#ifdef TRACE
		fprintf(stderr, "[%d]=", i);
#endif
		if (isprint(buf[i])) {
		  fputc(buf[i], stderr);
		} else {
		  fprintf(stderr, "[%02x]", buf[i]);
		}
	      }
	      fprintf(stderr, ">\n");
	      break;
	    }
	    bufo = eol + 1;
	  }
	  /* no end of line, wait for more */
	  else {
	    state = 0;
	    /* something got processed previously, throw it away */
	    if (bufo > 0) {
#ifdef TRACE
	      fprintf(stderr, "do(): Moving <%d> from <%d> to <0>\n", buft - bufo, bufo);
#endif
	      memmove(buf, buf + bufo, buft - bufo);
	      buft = buft - bufo;
	      bufo = 0;
	    }
	    /* nothing processed and line too long */
	    else if (buft > C_BUFSIZE / 2) {
	      fprintf(stderr, "Input line too long\n");
	      buft = 0;
	      bufo = 0;
	    }
	  }
	}
	/* nothing left to process */
	else {
	  buft = 0;
	  bufo = 0;
	  state = 0;
	}
      }

    } while (run);

    if (play) {
      w = c_clearend(&h);
      while (w > 0) {
	c_playout(dfd, &h);
	w--;
      }
    }
#ifdef TRACE
    c_dumpwave(&h, "-");
#endif

  }
  if (dfd != (-1)) {
    close(dfd);
  }
  return 0;
}

/* 21J.gf83Iap.Q */
